"""
CH E 210D Final Project
Isaac Zakaria
11 December 2019

"""
#import python modules
import numpy as np
import time
import multiprocessing as mp
import loky
import joblib
#import atomwrite
import matplotlib.pyplot as plt

#import compiled fortran library
import finallib
#import python library
import final

import sys
import os

def MDRunWrapper((M,lambdaD,RandomSeed)):
    
    # in the infinitely screened limit, kill the Coulomb term
    if lambdaD == 0:
        final.epsilonC = 0
        final.lambdaD = 1
    else:
        final.lambdaD = lambdaD
        final.epsilonC = 10
        
    final.M = M
    np.random.seed(RandomSeed)
    
    # non-global simulation variables
    rho = 0.8
    N = 300
    T = 1.0
    dt = 0.001
    NStepsEquil = 25000
    NStepsProd = 200000
    # global variables
    final.L = np.cbrt(N/rho)
    final.Cut = 2.5
    final.qList = (0,0,1,0,0,-1)*50
    
    Pos0 = final.InitPositions(N, final.L)
    PEnergies, AvgAtomCoord, MSDMatrix, DCoeff = final.MDRun(Pos0, N, T, rho, dt, NStepsEquil, NStepsProd)
    
    return M, lambdaD, PEnergies, AvgAtomCoord, MSDMatrix, DCoeff

def AvgStd(SimList, div=3):
    arrReshaped = SimList.reshape(-1,div)
    arrMeans = np.mean(arrReshaped, axis=1)
    arrStds = np.std(arrReshaped, axis=1)
    arrMeansStds = np.transpose(np.array((arrMeans, arrStds)))
    return arrMeansStds

ChainLengths = [3,6,12,15] # number of monomers per chain
lambdaDList = (1,5,10,15,20,30,40,50,75,100) # charge screening
SeedList = [342324,1337,420]
NumChains = len(ChainLengths)
NumLambdas = len(lambdaDList)
NumSeeds = len(SeedList)

# generate unique M, lambdaD pairs
argList = ()

for M in ChainLengths:
    for lambdaD in lambdaDList:
        for RandomSeed in SeedList:
            argList = argList + ((M, lambdaD, RandomSeed),)



#executor = loky.get_reusable_executor(timeout = 200, kill_workers = True)
#A = executor.map(MDRunWrapper, argList)

# set number of workers for multiprocessing
numWorkers = mp.cpu_count()

#if __name__ == '__main__':
#    A = joblib.Parallel(n_jobs = -1)(joblib.delayed(MDRunWrapper)(SimArgs) for SimArgs in argList)

flag = 0

# use multiprocessing to delegate one iteration to each available worker
if __name__ == '__main__':
    p = mp.Pool(processes = numWorkers)
    print "mp pool started"
    T0 = time.time()
    A = p.map_async(MDRunWrapper, argList).get(None)
    p.close()
    p.join()
    T1 = time.time()
    print "mp done; elapsed time in minutes..."
    print (T1-T0)/60
    flag = 1

if flag == 1:
    # initialize lists for each property in each simulation
    SimMList = []
    SimLambdaDList = []
    SimAbsERSDList = [] # absolute values of relative standard deviations of PEnergy
    SimAllAvgCoord = [] # average coordinations for all atoms
    SimAvgNegCoord = []
    SimAvgPosCoord = []
    SimAvgNeutralCoord = []
    SimAvgChargedCoord = []
    SimDCoeff = []
    
    qCharged = (0,0,1,0,0,1)*50
    qPos = (0,0,1,0,0,0)*50
    qNeg = (0,0,0,0,0,1)*50
    qNeutral = (1,1,0,1,1,0)*50
    
    for i,sim in enumerate(A):
        # unpack
        M, lambdaD, PEnergies, AvgAtomCoord, MSDMatrix, DCoeff = sim
        # collect M
        SimMList = np.append(SimMList, M)
        # collect lambdaD
        SimLambdaDList = np.append(SimLambdaDList, lambdaD)
        # collect absolute RSD of energy
        SimAbsERSDList = np.append(SimAbsERSDList, np.abs(np.std(PEnergies)/np.average(PEnergies)))
        # calculate average for all atoms
        SimAllAvgCoord = np.append(SimAllAvgCoord, np.average(AvgAtomCoord))
        # populate lists for average atom coordination
        AvgChargedCoord = qCharged * AvgAtomCoord
        AvgPosCoord = qPos * AvgAtomCoord
        AvgNegCoord = qNeg * AvgAtomCoord
        AvgNeutralCoord = qNeutral * AvgAtomCoord
        AvgChargedCoord = AvgChargedCoord[AvgChargedCoord != 0]
        AvgPosCoord = AvgPosCoord[AvgPosCoord != 0]
        AvgNegCoord = AvgNegCoord[AvgNegCoord != 0]
        AvgNeutralCoord = AvgNeutralCoord[AvgNeutralCoord != 0]
    #    for j,Atom in enumerate(AvgAtomCoord):
    #        AvgNegCoord = []
    #        AvgPosCoord = []
    #        AvgNeutralCoord = []
    #        # calculate averages separated by charge state
    #        if (j+1) % 3 == 0:
    #            if (j+1) % 2 == 0: # negative site
    #                AvgNegCoord = np.append(AvgNegCoord, Atom)
    #            else: # positive site
    #                AvgPosCoord = np.append(AvgPosCoord, Atom)
    #        else: # neutral site
    #            AvgNeutralCoord = np.append(AvgNeutralCoord, Atom)
        SimAvgChargedCoord = np.append(SimAvgChargedCoord, np.average(AvgChargedCoord))
        SimAvgNegCoord = np.append(SimAvgNegCoord, np.average(AvgNegCoord))
        SimAvgPosCoord = np.append(SimAvgPosCoord, np.average(AvgPosCoord))
        SimAvgNeutralCoord = np.append(SimAvgNeutralCoord, np.average(AvgNeutralCoord))
        # collect diffusion coefficients
        SimDCoeff = np.append(SimDCoeff, DCoeff)
    
    SimMList_MeanStd = AvgStd(SimMList)[:,0]
    SimLambdaDList_MeanStd = AvgStd(SimLambdaDList)[:,0]
    SimAbsERSDList_MeanStd = AvgStd(SimAbsERSDList)
    SimAllAvgCoord_MeanStd = AvgStd(SimAllAvgCoord)
    SimAvgChargedCoord_MeanStd = AvgStd(SimAvgChargedCoord)
    SimAvgPosCoord_MeanStd = AvgStd(SimAvgPosCoord)
    SimAvgNegCoord_MeanStd = AvgStd(SimAvgNegCoord)
    SimAvgNeutralCoord_MeanStd = AvgStd(SimAvgNeutralCoord)
    SimDCoeff_MeanStd = AvgStd(SimDCoeff)
    
    np.savetxt('SimMList_MeanStd3.txt',SimMList_MeanStd)
    np.savetxt('SimLambdaDList_MeanStd3.txt',SimLambdaDList_MeanStd)
    np.savetxt('SimAbsERSDList_MeanStd3.txt',SimAbsERSDList_MeanStd)
    np.savetxt('SimAllAvgCoord_MeanStd3.txt',SimAllAvgCoord_MeanStd)
    np.savetxt('SimAvgChargedCoord_MeanStd3.txt',SimAvgChargedCoord_MeanStd)
    np.savetxt('SimAvgNeutralCoord_MeanStd3.txt',SimAvgNeutralCoord_MeanStd)
    np.savetxt('SimDCoeff_MeanStd3.txt',SimDCoeff_MeanStd)